/* 1640, Wed 7 Jun 95

   METER_UX.C:  The AU Internet Accouting Meter mainline

   Copyright (C) 1992-1994 by Nevil Brownlee,
   Computer Centre,  The University of Auckland */

/*
 * snmpd.c - send snmp GET requests to a network entity.
 *
 */
/***********************************************************
	Copyright 1988, 1989 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#define noUX_TESTING

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/utsname.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>

#include <sys/time.h>

#ifdef SUNOS
#include <stropts.h>
#define void
#define const
#endif

#include "ausnmp.h"
#include "snmp.h"
#include "asn1.h"
#include "snmpimpl.h"

#define PKTSNAP
#include "pktsnap.h"
#include "flowhash.h"

#define APPLETALK

extern char version_descr[];  /* In met_vars.c */
extern unsigned int SampleRate;
extern char *communities[];  /* In snmp/snmpagnt.c */

#ifdef SOLARIS23
#define SNAPSIZE   54L  /* Enough for all except CLNS (which needs 70) */

/* SNAP size currently set to 54 to sidestep a bug in Solaris 2.3, 
   which corrupts packets smaller than the SNAP size.  This bug is 
   fixed in Solaris 2.4; if you are using Solaris 2.4 or you should
   set SNAPSIZE to MXPKTHDRLEN.   Nevil, 12 Jul 94 */

#else
#define SNAPSIZE   70L  /* Enough for CLNS */
#endif

/* ===== Interface code using libpcap to get see packet headers ===== */

#include <unistd.h>

#include <pcap-int.h>
#include <pcap.h>

pcap_t *pd = NULL;

unsigned long last_t_sec = 0;

struct ether_hdr {  
   unsigned char dest[6];    /* MAC addresses */
   unsigned char source[6];
   unsigned char type[2];    /* Length or protocol type */
   };

#include <llc.h>

/* ----- Following fddi code is from tcpdump's print-fddi.c ----- */

#ifdef __GNUC__  /* These defines from interface.h .. */
#define inline __inline
#ifndef __dead
#define __dead volatile
#endif
#else
#define inline
#define __dead
#endif

#include <fddi.h>

int	fddipad = FDDIPAD;	/* for proper alignment of header */

/*
 * Some FDDI interfaces use bit-swapped addresses.
 */
#if defined(ultrix) || defined(__alpha)
int	fddi_bitswap = 0;
#else
int	fddi_bitswap = 1;
#endif

/*
 * FDDI support for tcpdump, by Jeffrey Mogul [DECWRL], June 1992
 *
 * Based in part on code by Van Jacobson, which bears this note:
 *
 * NOTE:  This is a very preliminary hack for FDDI support.
 * There are all sorts of wired in constants & nothing (yet)
 * to print SMT packets as anything other than hex dumps.
 * Most of the necessary changes are waiting on my redoing
 * the "header" that a kernel fddi driver supplies to bpf:  I
 * want it to look like one byte of 'direction' (0 or 1
 * depending on whether the packet was inbound or outbound),
 * two bytes of system/driver dependent data (anything an
 * implementor thinks would be useful to filter on and/or
 * save per-packet, then the real 21-byte FDDI header.
 * Steve McCanne & I have also talked about adding the
 * 'direction' byte to all bpf headers (e.g., in the two
 * bytes of padding on an ethernet header).  It's not clear
 * we could do this in a backwards compatible way & we hate
 * the idea of an incompatible bpf change.  Discussions are
 * proceeding.
 *
 * Also, to really support FDDI (and better support 802.2
 * over ethernet) we really need to re-think the rather simple
 * minded assumptions about fixed length & fixed format link
 * level headers made in gencode.c.  One day...
 *
 *  - vj
 */

#define FDDI_HDRLEN (sizeof(struct fddi_header))

static u_char fddi_bit_swap[] = {
	0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,
	0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
	0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,
	0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
	0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,
	0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
	0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,
	0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
	0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,
	0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
	0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,
	0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
	0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,
	0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
	0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,
	0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
	0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,
	0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
	0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,
	0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
	0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,
	0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
	0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,
	0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
	0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,
	0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
	0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,
	0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
	0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,
	0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
	0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,
	0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff,
};

/* Extract src, dst addresses */
static inline void extract_fddi_addrs(
   const struct fddi_header *fddip, char *fsrc, char *fdst)
{
	register int i;

	if (fddi_bitswap) {
		/*
		 * bit-swap the fddi addresses (isn't the IEEE standards
		 * process wonderful!) then convert them to names.
		 */
		for (i = 0; i < 6; ++i)
			fdst[i] = fddi_bit_swap[fddip->fddi_dhost[i]];
		for (i = 0; i < 6; ++i)
			fsrc[i] = fddi_bit_swap[fddip->fddi_shost[i]];
	}
	else {
		memcpy(fdst, fddip->fddi_dhost, 6);
	        memcpy(fsrc, fddip->fddi_shost, 6);
	}
}

/* ----- End of fddi code from tcpdump's print-fddi.c ----- */

void fddi_callback(u_char *user, struct pcap_pkthdr *h, u_char *p)
{
   const struct fddi_header *fp = (struct fddi_header *)p;
   int caplen = h->caplen;
   struct ether_hdr ehdr;
   struct llc llc;
   struct pkt pp;
   unsigned int ether_type, lsap;

   if (caplen < FDDI_HDRLEN) return;  /* Too short! */

   /* Get the FDDI addresses into a canonical form */
   extract_fddi_addrs(fp, ehdr.source, ehdr.dest);

   if ((fp->fddi_fc & FDDIFC_CLFF) == FDDIFC_LLC_ASYNC) {
      caplen -= FDDI_HDRLEN;
      if (caplen < sizeof(struct llc)) return;  /* Too short! */
      p += FDDI_HDRLEN;
      memcpy((char *)&llc, (char *)p, sizeof(llc));
      lsap = llc.dsap << 8 | llc.ssap;
      if (lsap == 0xAAAA) {  /* SNAP packet */
         ether_type = GET_SHORT(llc.ctl.snap_ether.snap_ethertype);
         p += sizeof(struct llc);
         }
      else ether_type = 0;  /* Other 802.2 packet */
      }
   handle_pkt(&pp, ether_type,lsap, h->len, (unsigned char *)&ehdr,p);
   pkt_monitor(&pp);
   if (h->ts.tv_sec != last_t_sec) {
      last_t_sec = h->ts.tv_sec;
      one_second_process();
      }
   }

#ifdef  XXYYZZ
unsigned int line_nbr=0;
#endif

void ether_callback(u_char *user, struct pcap_pkthdr *h, u_char *p)
{
   struct ether_hdr *ethp;
   struct llc *llcp;
   struct pkt pp;
   unsigned int ether_type, lsap;
   int len, j;

   ethp = (struct ether_hdr *)p;
   ether_type = GET_SHORT(ethp->type);
   p += sizeof(struct ether_hdr);
   if (ether_type <= 1500) {  /* 802.3 packet */
      llcp = (struct llc *)p;
      lsap = llcp->dsap << 8 | llcp->ssap;
      if (lsap == 0xAAAA) {  /* SNAP packet */
         ether_type = GET_SHORT(llcp->ctl.snap_ether.snap_ethertype);
         p += sizeof(struct llc);
         }
      else ether_type = 0;  /* Other 802.2 packet */
      }
   else lsap = 0;  /* Blue book packet */

#ifdef  XXYYZZ
/*   if (memcmp("\x00\x00\xA7\x11\xA7\xF3", ep+6, 6) == 0) { */
   if (ether_type == 0x0800) {
      unsigned char *ep=(unsigned char *)ethp;
      fprintf(stderr, "%5d %5d %4d:%03d.%03d  %02X-%02X-%02X-%02X-%02X-%02X  ",
         ++line_nbr, h->len,
         h->ts.tv_sec%1000, h->ts.tv_usec/1000, h->ts.tv_usec%1000,
         ep[6],ep[7],ep[8],ep[9],ep[10],ep[11]);
      fprintf(stderr, "%02X-%02X-%02X-%02X-%02X-%02X  %02X",
         ep[0],ep[1],ep[2],ep[3],ep[4],ep[5], ep[12]);
      for (j=13; j!=53; ++j) fprintf(stderr,"-%02X", ep[j]);
      fprintf(stderr,"\n");
/*      printf("\r%6d", ++line_nbr); */
      }
#endif

   handle_pkt(&pp, ether_type,lsap, h->len, (unsigned char *)ethp,p);
   pkt_monitor(&pp);
   if (h->ts.tv_sec != last_t_sec) {
      last_t_sec = h->ts.tv_sec;
      one_second_process();
      }
   }

pcap_handler callback;

void interface_read(int fd)
{
   pcap_read(pd, -1, callback, (u_char *)NULL);
   }

int init_interface(char *device)
{
   char *interface, errbuf[PCAP_ERRBUF_SIZE];
   int type;

   if (*device == '\0')
      if (!(interface = pcap_lookupdev(errbuf))) {
         printf("pcap_lookupdevice(): %s\n",errbuf);
         return -1;
         }
      else strcpy(device,interface);
   if ((pd = pcap_open_live(device, SNAPSIZE, 1, 250, errbuf)) == NULL) {
      printf("pcap_open_live(%s): %s\n", device,errbuf);
      return -1;
      }
   setuid(getuid());

   type = pcap_datalink(pd);
   if (type == DLT_EN10MB) {
      AdjAddrType = 1;
      callback = (pcap_handler)ether_callback;
      }
   else if (type == DLT_FDDI) {
      AdjAddrType = 8;      callback = (pcap_handler)fddi_callback;
      }
   else {   
      printf("pcap bad link type 0x%x!\n", type);
      pcap_close(pd);
      return -1;
      }
   return pd->fd;  /* Actual pcap socket */
   }

unsigned int pcap_lost_packets(void)
{
   struct pcap_stat ps;
   char errbuf[PCAP_ERRBUF_SIZE];
   if (pcap_stats(pd,&ps)) {
      printf("pcap_stats(): %s\n",errbuf);
      return 0;
      }
   return ps.ps_drop;
   } 

/* ===== End of libpcap interface code ===== */

time_t accounting_start_time;  /* Seconds since start of Unix epoch */

unsigned long uptime(void)  /* SNMP time */
{
   return (time(NULL)-accounting_start_time)*100;  /* centiseconds */
   }

void show_meter_time()
{
   char msg[60], *ts;
   ts = ctime(&accounting_start_time);
   sprintf(msg,"%lu seconds since %c%c%c%c:%c%c",
      time(NULL)-accounting_start_time,
      ts[11],ts[12],ts[14],ts[15], ts[17],ts[18]);
   display_msg(0,msg);
   }

extern int errno;
int snmp_dump_packet = 0;

char interface_name[32] = "";  /* Device name for pcap interface */

void main(int argc, char *argv[])
{
   int arg,c, sd,if_fd;
   char *ap;
   struct sockaddr_in me;
   struct utsname name;

   printf("%s\n",version_descr);
   display_enabled = kb_enabled = 1;  /* Enabled by default */
   mxflowsp1 = DFMXFLOWS+1;
   accounting_start_time = time(NULL) - 2;  /* So flows have CreateTime > 1 */
   /* Default (CMU) communities are:  0 = "public", 1 = "proxy",
	 2 = "private", 3 = "regional", 4 = "core"
      We only allow "public" and "private" by default */
   communities[1] = communities[3] = communities[4] = "";
   for (c = 0, arg = 1; arg < argc; ++arg) {
      if (argv[arg][0] == '-') {
	 switch (argv[arg][1]) {
	 case 'd':
	    snmp_dump_packet++;
	    break;
	 case 'f':
	    ap = argv[arg]+2;
	    if (*ap == '\0') ap = argv[++arg];
	    mxflowsp1 = atoi(ap)+1;
	    break;
	 case 'i':  /* -i name of interface to monitor */
	    strcpy(interface_name,argv[++arg]);
	    break;
	 case 'k':
	    kb_enabled = 0;  /* -k to disable keyboard */
	    break;
 	 case 'n':
 	    ap = argv[arg]+2;
 	    if (*ap == '\0') ap = argv[++arg];
 	    SampleRate = atoi(ap);
 	    if (SampleRate < 1) SampleRate = 1;
 	    break;
	 case 'r':
	    for (;;) {
	       if (c == 5) {
                  printf("Max of 4 read communities allowed\n");
                  exit(0);
                  }
	       communities[c++] = argv[++arg];
	       if (c == 2) ++c;  /* 2 -> "private" */
	       if (arg == argc-1 || argv[arg+1][0] == '-') break;
               }
	    break;
	 case 's':
	    display_enabled = 0;  /* -s to disable screen */
	    break;
	 case 'w':
	    communities[2] = argv[++arg];  /* -w to set write community */
	    break;
	 default:
	    printf("Invalid option: -%c\n", argv[arg][1]);
	    exit(0);
	    }
	 }
      }
   /* Set up connections */
   sd = socket(AF_INET, SOCK_DGRAM, 0);
   if (sd < 0){
      perror("socket");
      exit(1);
      }
   me.sin_family = AF_INET;
   me.sin_addr.s_addr = INADDR_ANY;
   me.sin_port = htons(SNMP_PORT);
   if (bind(sd, (struct sockaddr *)&me, sizeof(me)) != 0) {
      perror("bind");
      exit(2);
      }
   init_monitor();
   init_snmp();
   if_fd = init_interface(interface_name);
   if (if_fd < 0) exit(3);
   uname(&name);
   printf("Running on %s, interface %s\n",
      name.nodename, interface_name);
   receive(sd,if_fd);
   }

void zero_pkt_stats()
{
   kilodummypackets = npackets = lostpackets = stats_time = spackets = 0L;
   max_pkt_rate = 0;
   clear_pkt_stats = 0;
   }

#define BKG_INTERVAL  30  /* Seconds */
int gci;

void one_second_process(void)
{
   unsigned long pd;
   unsigned char p;
   s_uptime = uptime();  /* Don't do this too often! */
#ifdef UX_TESTING
   printf(".");
#endif
   if (clear_pkt_stats) zero_pkt_stats();
   else {
      ++stats_time;
      pd = (npackets-spackets)*SampleRate;
      if (pd > max_pkt_rate) max_pkt_rate = pd;
      spackets = npackets;
      }
   if (--gci == 0) {
#ifdef UX_TESTING
     printf("g");
#endif
      garbage_collect(1);  /* Routine incremental collection */
      gci = gc_interval;
      }
   if (--bkgi == 0) {  /* Check % of flows active */
#ifdef UX_TESTING
      printf("b\n");
#endif
      p = (unsigned long)active_flows()*100/(mxflowsp1-1);
      if (p > FloodMark && FloodMark > 0 && FloodMark < 100) {
         if (CurrentRuleSet != 1) {
	    close_rule_set();
	    open_rule_set(1,1);
	    display_msg(1,"Switched to default rules");
	    }
	 }
      else if (p > HighWaterMark &&
	    HighWaterMark > 0 && HighWaterMark < 100 &&
	    EmergencyRuleSet != 0 &&
	    CurrentRuleSet != EmergencyRuleSet) {
         close_rule_set();
	 open_rule_set(1,EmergencyRuleSet);
	 display_msg(1,"Switched to emergency rules");
	 }
      else if (p > HighWaterMark*2/3)
	 more_garbage();
      bkgi = BKG_INTERVAL;
      }
#ifdef UX_TESTING
   fflush(stdout);
#endif
   }

handle_keyboard()
{
   char kb_buf[25];
   int ch;
   fgets(kb_buf, sizeof kb_buf, stdin);
   if ((ch = kb_buf[0]) == 27) {  /* ESC */
      display_msg(1,"Shutting down");
      exit(0);
      }
   switch (tolower(ch)) {
   case 'v':
      printf("\n%s\n",version_descr);
      break;
   default:
      handle_kb(ch);
      break;
      }
   }

receive(sd,int_fd)
    int sd,int_fd;
{
   fd_set fdset;
   int count;
   struct timeval t;

   bkgi = BKG_INTERVAL;
   zero_pkt_stats();
   t.tv_sec = 0;  t.tv_usec = 250000;  /* 0.25 seconds */
   gci = gc_interval;  /* Garbage collect interval */
   bkgi = BKG_INTERVAL;  /* Background interval */
   gc_f = (mxflowsp1-1)/100;  /* Flow indexes to check */
   if (gc_f < 4) gc_f = 4;
   s_uptime = uptime();  /* Initial value */

   for (;;) {
      FD_ZERO(&fdset);
      FD_SET(int_fd,&fdset);
      FD_SET(sd, &fdset);
      if (kb_enabled) FD_SET(0, &fdset);  /* stdin */

      count = select(FD_SETSIZE, &fdset, 0, 0, &t);
      if (count > 0) {
         if (FD_ISSET(int_fd, &fdset))
            interface_read(int_fd);
         if (FD_ISSET(sd, &fdset))
	    snmp_read(sd);
         if (FD_ISSET(0, &fdset))  /* stdin */
            handle_keyboard();
         }
      else switch (count) {
      case 0:
	 break;
      case -1:
	 if (errno == EINTR) continue;
	 else perror("select");
	 return -1;
      default:
	 printf("select returned %d\n", count);
	 return -1;
	 }
      }
   }

unsigned long snmp_peer_addr;

snmp_read(int sd)
{
   struct sockaddr_in from;
   int length, out_length, fromlength, count;
   char  packet[1500], outpacket[1500];
   char snmp_peer_name[50];  /* Name of host which sent the snmp request */

   fromlength = sizeof from;
   length = recvfrom(sd, packet, 1500, 0, 
      (struct sockaddr *)&from, &fromlength);
   snmp_peer_addr = from.sin_addr.s_addr;
   strcpy(snmp_peer_name,inet_ntoa(from.sin_addr));
   if (length == -1) perror("recvfrom");
   if (snmp_dump_packet) {
      printf("received %d bytes from %s:\n", length, snmp_peer_name);
      for (count = 0; count < length; count++) {
         printf("%02X ", packet[count]);
	 if ((count % 16) == 15) printf("\n");
	 }
      printf("\n\n");
      }

   out_length = 1500;
   if (snmp_agent_parse(packet, length, outpacket, &out_length, 
         from.sin_addr.s_addr)) {
      if (snmp_dump_packet) {
	 printf("sent %d bytes to %s:\n", out_length, snmp_peer_name);
	 for (count = 0; count < out_length; count++) {
            printf("%02X ", outpacket[count]);
	    if ((count % 16) == 15) printf("\n");
	    }
	    printf("\n\n");
	 }
      if (sendto(sd, (char *)outpacket, out_length, 0, 
            (struct sockaddr *)&from, sizeof(from)) < 0) {
         perror("sendto");
	 return 0;
	 }
      }
   }

